import csv
import logging
import re
import sys
import tempfile
import datetime
from datetime import datetime as dt
from io import StringIO
from pathlib import Path
from time import sleep

import pytz
from django.db import models, transaction

from preservationdatabase import utils

LIMIT_TIME_DELTA = datetime.timedelta(weeks=3)


class LastFill(models.Model):
    class Meta:
        db_table = "preservationData_lastfill"
        app_label = "preservationdatabase"

    archive_name = models.CharField(max_length=255)
    last_fill_date = models.DateTimeField(null=True, blank=True)

    @staticmethod
    def set_last_fill(archive_name, last_fill_date=dt.now(pytz.utc)):
        lf, created = LastFill.objects.get_or_create(archive_name=archive_name)
        lf.last_fill_date = last_fill_date
        lf.save()

    @staticmethod
    def clear():
        LastFill.objects.all().delete()

    @staticmethod
    def get_last_fill(archive_name):
        try:
            return LastFill.objects.get(
                archive_name=archive_name
            ).last_fill_date
        except LastFill.DoesNotExist:
            return None

    def cache_valid(self) -> bool:
        time_delta = datetime.datetime.now(pytz.utc) - self.last_fill_date

        return True if LIMIT_TIME_DELTA > time_delta else False

    @staticmethod
    def will_cache(archive_name) -> bool:
        try:
            lf = LastFill.objects.get(archive_name=archive_name)
            return lf.cache_valid()
        except LastFill.DoesNotExist:
            return False

    def __str__(self):
        return "{}: {}".format(self.archive_name, self.last_fill_date)


class Publisher(models.Model):
    class Meta:
        db_table = "preservationData_publisher"
        app_label = "preservationdatabase"

    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name


class LockssPreservation(models.Model):
    class Meta:
        db_table = "preservationData_locksspreservation"

        indexes = [
            models.Index(fields=["issn"]),
            models.Index(fields=["eissn"]),
            models.Index(fields=["title"]),
        ]

    title = models.TextField()
    issn = models.CharField(max_length=20)
    eissn = models.CharField(max_length=20)
    preserved_volumes = models.TextField()
    preserved_years = models.TextField()
    in_progress_volumes = models.TextField()
    in_progress_years = models.TextField()
    publisher = models.ForeignKey(
        Publisher, on_delete=models.CASCADE, default=None
    )

    def __str__(self):
        return (
            "{} (issn: {}) (eissn: {}) (v{}({})) "
            "(in progress: {} ({})) ".format(
                self.title,
                self.issn,
                self.eissn,
                self.preserved_volumes,
                self.preserved_years,
                self.in_progress_volumes,
                self.in_progress_years,
            )
        )

    @staticmethod
    def name() -> str:
        return "LOCKSS"

    @staticmethod
    def preservation(
        container_title: str,
        issn: str,
        volume: str,
        no=None,
        year=None,
        doi=None,
    ):
        """
        Determine whether a DOI is preserved in LOCKSS
        :param container_title: the container title
        :param issn: the ISSN
        :param volume: the volume
        :param no: the issue number
        :param year: the year
        :param doi: a DOI
        :return: A LockssPreservation item (or None) and a bool indicating
        whether the item is fully preserved
        """
        return utils.preservation_status(
            LockssPreservation, container_title, issn, volume, no=no, year=year
        )

    @staticmethod
    def create_preservation(
        issn,
        eissn,
        title,
        preserved_volumes,
        preserved_years,
        in_progress_volumes,
        in_progress_years,
        publisher,
        model,
    ) -> None:
        """
        Create a preservation item of this model
        :param issn: the ISSN
        :param eissn: the eISSN
        :param title: the title
        :param preserved_volumes: the preserved volumes
        :param preserved_years: the preserved years
        :param in_progress_volumes: the in-progress volumes
        :param in_progress_years: the in-progress years
        :param publisher: the publisher object
        :param model: the model on which to operate
        :return: None
        """
        model.objects.create(
            issn=issn,
            eissn=eissn,
            title=title,
            preserved_volumes=preserved_volumes,
            preserved_years=preserved_years,
            in_progress_volumes=in_progress_volumes,
            in_progress_years=in_progress_years,
            publisher=publisher,
        )

    @staticmethod
    @transaction.atomic
    def import_data(url: str = None, local: bool = False) -> None:
        """
        Import data into the system
        :param url: the URL of the data file
        :param local: whether the data file is local
        :return: None
        """
        utils.generic_lockss_import(
            url, LockssPreservation, local=local, skip_first_line=True
        )


class CarinianaPreservation(models.Model):
    class Meta:
        db_table = "preservationData_carinianapreservation"

        indexes = [
            models.Index(fields=["issn"]),
            models.Index(fields=["eissn"]),
            models.Index(fields=["title"]),
        ]

    title = models.TextField()
    issn = models.CharField(max_length=20)
    eissn = models.CharField(max_length=20)
    preserved_volumes = models.TextField()
    preserved_years = models.TextField()
    in_progress_volumes = models.TextField()
    in_progress_years = models.TextField()
    publisher = models.ForeignKey(
        Publisher, on_delete=models.CASCADE, default=None
    )

    def __str__(self):
        return (
            "{} (issn: {}) (eissn: {}) (v{}({})) "
            "(in progress: {} ({})) ".format(
                self.title,
                self.issn,
                self.eissn,
                self.preserved_volumes,
                self.preserved_years,
                self.in_progress_volumes,
                self.in_progress_years,
            )
        )

    @staticmethod
    def name() -> str:
        return "Cariniana"

    @staticmethod
    def preservation(
        container_title: str,
        issn: str,
        volume: str,
        no=None,
        year=None,
        doi=None,
    ):
        """
        Determine whether a DOI is preserved in Cariniana
        :param container_title: the container title
        :param issn: the ISSN
        :param volume: the volume
        :param no: the issue number
        :param year: the year
        :param doi: a DOI
        :return: A Cariniana item (or None) and a bool indicating
        whether the item is fully preserved
        """
        return utils.preservation_status(
            CarinianaPreservation,
            container_title,
            issn,
            volume,
            no=no,
            year=year,
        )

    @staticmethod
    def create_preservation(
        issn,
        eissn,
        title,
        preserved_volumes,
        preserved_years,
        in_progress_volumes,
        in_progress_years,
        publisher,
        model,
    ) -> None:
        """
        Create a preservation item of this model
        :param issn: the ISSN
        :param eissn: the eISSN
        :param title: the title
        :param preserved_volumes: the preserved volumes
        :param preserved_years: the preserved years
        :param in_progress_volumes: the in-progress volumes
        :param in_progress_years: the in-progress years
        :param publisher: the publisher object
        :param model: the model on which to operate
        :return: None
        """
        model.objects.create(
            issn=issn,
            eissn=eissn,
            title=title,
            preserved_volumes=preserved_volumes,
            preserved_years=preserved_years,
            in_progress_volumes=in_progress_volumes,
            in_progress_years=in_progress_years,
            publisher=publisher,
        )

    @staticmethod
    @transaction.atomic
    def import_data(url: str = None, local: bool = False) -> None:
        """
        Import data into the system
        :param url: the URL of the data file
        :param local: whether the data file is local
        :return: None
        """
        utils.generic_lockss_import(
            url, CarinianaPreservation, local=local, skip_first_line=True
        )


class ClockssPreservation(models.Model):
    class Meta:
        db_table = "preservationData_clocksspreservation"

        indexes = [
            models.Index(fields=["issn"]),
            models.Index(fields=["eissn"]),
            models.Index(fields=["title"]),
        ]

    title = models.TextField()
    issn = models.CharField(max_length=20)
    eissn = models.CharField(max_length=20)
    preserved_volumes = models.TextField()
    preserved_years = models.TextField()
    in_progress_volumes = models.TextField()
    in_progress_years = models.TextField()
    publisher = models.ForeignKey(
        Publisher, on_delete=models.CASCADE, default=None
    )

    def __str__(self):
        return (
            "{} (issn: {}) (eissn: {}) (v{}({})) "
            "(in progress: {} ({})) ".format(
                self.title,
                self.issn,
                self.eissn,
                self.preserved_volumes,
                self.preserved_years,
                self.in_progress_volumes,
                self.in_progress_years,
            )
        )

    @staticmethod
    def name() -> str:
        return "CLOCKSS"

    @staticmethod
    def preservation(
        container_title: str,
        issn: str,
        volume: str,
        no=None,
        year=None,
        doi=None,
    ):
        """
        Determine whether a DOI is preserved in CLOCKSS
        :param container_title: the container title
        :param issn: the ISSN
        :param volume: the volume
        :param no: the issue number
        :param year: the year
        :param doi: a DOI
        :return: A ClockssPreservation item (or None) and a bool indicating
        whether the item is fully preserved
        """
        return utils.preservation_status(
            ClockssPreservation, container_title, issn, volume, no=no, year=year
        )

    @staticmethod
    def create_preservation(
        issn,
        eissn,
        title,
        preserved_volumes,
        preserved_years,
        in_progress_volumes,
        in_progress_years,
        publisher,
        model,
    ) -> None:
        """
        Create a preservation item of this model
        :param issn: the ISSN
        :param eissn: the eISSN
        :param title: the title
        :param preserved_volumes: the preserved volumes
        :param preserved_years: the preserved years
        :param in_progress_volumes: the in-progress volumes
        :param in_progress_years: the in-progress years
        :param publisher: the publisher object
        :param model: the model on which to operate
        :return: None
        """
        model.objects.create(
            issn=issn,
            eissn=eissn,
            title=title,
            preserved_volumes=preserved_volumes,
            preserved_years=preserved_years,
            in_progress_volumes=in_progress_volumes,
            in_progress_years=in_progress_years,
            publisher=publisher,
        )

    @staticmethod
    @transaction.atomic
    def import_data(url: str = None, local: bool = False) -> None:
        """
        Import data into the system
        :param url: the URL of the data file
        :param local: whether the data file is local
        :return: None
        """
        utils.generic_lockss_import(
            url, ClockssPreservation, local=local, skip_first_line=True
        )


class PKPPreservation(models.Model):
    class Meta:
        db_table = "preservationData_pkppreservation"

        indexes = [
            models.Index(fields=["issn"]),
            models.Index(fields=["title"]),
        ]

    title = models.TextField()
    issn = models.CharField(max_length=20)
    preserved_volumes = models.TextField()
    preserved_no = models.TextField(blank=True, null=True)
    publisher = models.ForeignKey(
        Publisher, on_delete=models.CASCADE, default=None
    )

    def __str__(self):
        return "{} (issn: {}) (v{}({})) ".format(
            self.title,
            self.issn,
            self.preserved_volumes,
            self.preserved_no,
        )

    @staticmethod
    def name() -> str:
        return "PKP PLN"

    @staticmethod
    def preservation(container_title, issn, volume, no, year=None, doi=None):
        """
        Determine whether a DOI is preserved in the PKP private LOCKSS network
        :param container_title: the container title
        :param issn: the ISSN
        :param volume: the volume
        :param no: the issue number
        :param year: the year
        :param doi: a DOI
        :return: A PKPPreservation item (or None) and a bool indicating
        whether the item is fully preserved
        """
        preserved_item = utils.get_preserved_item_record(
            PKPPreservation, container_title, issn
        )

        if not preserved_item or len(preserved_item) == 0:
            return None, False

        if no is not None and no != "" and no != "0":
            preserved_item.filter(preserved_no=no)

            if len(preserved_item) == 0:
                return None, False

        return preserved_item, True

    @staticmethod
    @transaction.atomic
    def import_data(url: str = None, local: bool = False) -> None:
        """
        Import data into the system
        :param url: the URL of the data file
        :param local: whether the data file is local
        :return: None
        """
        utils.generic_lockss_import(
            url, PKPPreservation, local=local, skip_first_line=True
        )

    @staticmethod
    def create_preservation(
        issn, title, preserved_volumes, preserved_no, publisher, model
    ) -> None:
        """
        Create a preservation item of this model
        :param issn: the ISSN
        :param title: the title
        :param preserved_volumes: the preserved volumes
        :param publisher: the publisher object
        :param model: the model on which to operate
        :return: None
        """
        model.objects.create(
            issn=issn,
            title=title,
            preserved_volumes=preserved_volumes,
            preserved_no=preserved_no,
            publisher=publisher,
        )


class HathiPreservation(models.Model):
    class Meta:
        db_table = "preservationData_hathipreservation"

        indexes = [
            models.Index(fields=["issn"]),
            models.Index(fields=["title"]),
        ]

    title = models.TextField()
    issn = models.CharField(max_length=20)
    preserved_volumes = models.TextField()

    def __str__(self):
        return "{} (issn: {}) (v{}) ".format(
            self.title, self.issn, self.preserved_volumes
        )

    @staticmethod
    def name() -> str:
        return "HathiTrust"

    @staticmethod
    def preservation(
        container_title, issn, volume, no=None, year=None, doi=None
    ):
        """
        Determine whether a DOI is preserved in HathiTrust
        :param container_title: the container title
        :param issn: the ISSN
        :param volume: the volume
        :param no: the issue number
        :param year: the year
        :param doi: a DOI
        :return: A HathiPreservation item (or None) and a bool indicating
        whether the item is fully preserved
        """
        preserved_item = utils.get_preserved_item_record(
            HathiPreservation, container_title, issn
        )

        if not preserved_item or len(preserved_item) == 0:
            return None, False

        if preserved_item:
            # HathiTrust records volume strings as "v[1, 2]" etc
            # but also as "v1999" etc.

            if year:
                volume_year_regex = re.compile(rf"v{year}")
                volume_year_regex_two = re.compile(rf"{year}")

                for pi in preserved_item:
                    matches = re.findall(
                        volume_year_regex, pi.preserved_volumes
                    )

                    if matches:
                        return pi, True

                    matches = re.findall(
                        volume_year_regex_two, pi.preserved_volumes
                    )

                    if matches:
                        return pi, True

            volume_regex = r"(\d+),?"

            volume = str(volume)

            for pi in preserved_item:
                matches = re.finditer(
                    volume_regex, pi.preserved_volumes, re.MULTILINE
                )

                for matchNum, match in enumerate(matches, start=1):
                    for groupNum in range(0, len(match.groups())):
                        groupNum = groupNum + 1

                        group = match.group(groupNum)

                        if group == volume:
                            return pi, True

            return None, False

        else:
            return None, False

    @staticmethod
    @transaction.atomic
    def import_data(url, bucket="", s3client=None, local=False):
        # download the data file from S3 bucket
        with tempfile.TemporaryDirectory() as tmp:
            if local:
                path = url
            else:
                path = Path(tmp) / "downloaded.file"

                hathi_data = utils.download_remote(
                    False,
                    HathiPreservation,
                    url,
                    bucket=bucket,
                    s3client=s3client,
                    decode=False,
                    file=True,
                    filename=str(path),
                )

            # clear out existing data
            utils.clear_out(HathiPreservation)

            csv.field_size_limit(sys.maxsize)

            with open(str(path), "r") as input_file:
                csv_reader = csv.reader(input_file, delimiter="\t")

                volume_matcher = r"v\.\s?(\d+(?:\-?\d+)?)"
                no_matcher = r"no\.\s?(\d+(?:\-?\d+)?)"
                year_matcher = r"\d{4}"
                issn_matcher = r"[0-9][0-9][0-9][0-9][-][0-9][0-9][0-9][X0-9]"

                for row in csv_reader:
                    try:
                        vols = row[4]
                        issn = row[9]
                        issns = []
                        title = row[11]
                        publishing_info = row[12]
                        date = row[16]
                        bf = row[19]
                        unknown_format = False

                        # check if it's an ISSN
                        # logging.info('ISSN: {}'.format(issn))

                        # Sometimes we have formats like this for ISSN:
                        # 1938-1603 (online),0022-1864
                        issn_matches = re.findall(issn_matcher, issn)

                        if issn_matches:
                            for match in issn_matches:
                                issns.append(match)

                        # if it's a serial, try to parse the vols
                        if bf == "SE" and len(issns) > 0 and vols:
                            matches = re.findall(volume_matcher, vols)

                            if matches:
                                for match in matches:
                                    vols = utils.unpack_range(match)
                            else:
                                matches = re.findall(no_matcher, vols)

                                if matches:
                                    for match in matches:
                                        vols = utils.unpack_range(match)
                                else:
                                    matches = re.findall(year_matcher, vols)

                                    if matches:
                                        for match in matches:
                                            vols = match
                                    else:
                                        unknown_format = True

                            if not unknown_format:
                                if title.endswith("."):
                                    title = title[:-1]

                                for issn_no in issns:
                                    HathiPreservation.objects.create(
                                        issn=issn_no,
                                        title=title,
                                        preserved_volumes=vols,
                                    )

                                logging.info(
                                    f"Added {title} (issns {issns}) to "
                                    f"{HathiPreservation.name()} data"
                                )
                    except IndexError:
                        pass


class PorticoPreservation(models.Model):
    class Meta:
        db_table = "preservationData_porticopreservation"

        indexes = [
            models.Index(fields=["issn"]),
            models.Index(fields=["eissn"]),
            models.Index(fields=["title"]),
        ]

    title = models.TextField()
    issn = models.CharField(max_length=20)
    eissn = models.CharField(max_length=20)
    preserved_volumes = models.TextField()

    # this indicates whether the title is preserved or queued
    preserved = models.BooleanField()
    publisher = models.ForeignKey(
        Publisher, on_delete=models.CASCADE, default=None
    )

    def __str__(self):
        return "{} (issn: {}) (eissn: {}) (v{}) ".format(
            self.title, self.issn, self.eissn, self.preserved_volumes
        )

    @staticmethod
    def name():
        return "Portico"

    @staticmethod
    def preservation(
        container_title, issn, volume, no=None, year=None, doi=None
    ):
        """
        Determine whether a DOI is preserved in Portico
        :param container_title: the container title
        :param issn: the ISSN
        :param volume: the volume
        :param no: the issue number
        :param year: the year
        :param doi: a DOI
        :return: A PorticoPreservation item (or None) and a bool indicating
        whether the item is fully preserved
        """
        preserved_item = utils.get_preserved_item_record(
            PorticoPreservation, container_title, issn
        )

        if not preserved_item or len(preserved_item) == 0:
            return None, False

        # Portico gives volume formats as follows:
        # 2013/2014 - v. 2 (1-2)
        volume_regex = r"v\.\s(\d+)"

        year_regex = r"\d{4}"

        for pi in preserved_item:
            matches = re.findall(volume_regex, pi.preserved_volumes)

            volume = str(volume)

            if volume in matches:
                return pi, pi.preserved

            year_matches = re.findall(year_regex, pi.preserved_volumes)
            year = str(year)

            if year in year_matches:
                return pi, pi.preserved

        return None, False

    @staticmethod
    @transaction.atomic
    def import_data(url: str = None, local: bool = False) -> None:
        """
        Import data into the system
        :param url: the URL of the data file
        :param local: whether the data file is local
        :return: None
        """
        # get CSV data
        csv_file = utils.download_remote(local, PorticoPreservation, url)

        # clear out
        utils.clear_out(PorticoPreservation)

        # increase the CSV field size to accommodate large entries
        csv.field_size_limit(sys.maxsize)

        with StringIO(csv_file) as input_file:
            csv_reader = csv.DictReader(
                input_file, delimiter="\t", quoting=csv.QUOTE_NONE
            )

            for row in csv_reader:
                publisher, created = Publisher.objects.get_or_create(
                    name=row["publisher_name"]
                )

                PorticoPreservation.objects.create(
                    issn=row["print_identifier"],
                    eissn=row["online_identifier"],
                    title=row["publication_title"],
                    preserved_volumes=row["holding_list"],
                    preserved=(row["notes"] == "Preserved"),
                    publisher=publisher,
                )

                logging.info(
                    f'Added {row["publication_title"]} to '
                    f"{PorticoPreservation.name()} data"
                )


class OculScholarsPortalPreservation(models.Model):
    class Meta:
        db_table = "preservationData_oculpreservation"

        indexes = [
            models.Index(fields=["issn"]),
            models.Index(fields=["title"]),
        ]

    title = models.TextField()

    """
    ONIX field codes to identify an ISSN:
    ResourceVersionIDType = 7
    IDValue = unhyphenated ISSN
    
    e.g.
    <oph:ResourceVersion>
        <oph:ResourceVersionIdentifier>
        <oph:ResourceVersionIDType>07</oph:ResourceVersionIDType>
        <oph:IDValue> 27697541</oph:IDValue>
    </oph:ResourceVersionIdentifier>
    """
    issn = models.CharField(max_length=20)
    preserved_volumes = models.TextField()
    preserved_no = models.TextField()

    # this indicates whether the title is preserved or queued
    """
    ONIX field codes for the <oph:PreservationStatus><oph:PreservationStatusCode> tags:
    00	Unknown	The preservation status is genuinely unknown, and should be updated as soon as information is available	5		
    01	Will not be preserved	Preservation agency has decided against preservation of the issues and e-content in the <Coverage> statement	5		
    02	Undecided	Preservation agency is considering preservation of the issues and e-content in the <Coverage> statement. No decision has yet been made. This status should be updated as soon as information is available	5		
    03	Committed	Preservation agency is committed to preserving the issues and e-content in the <Coverage> statement. No active steps have yet been taken	5		
    04	In progress	Preservation agency is in the process of preserving the issues and e-content in the <Coverage> statement	5		
    05	Preserved	Preservation agency has preserved the issues and e-content in the <Coverage> statement	5
    """
    preserved = models.IntegerField()

    """
    ONIX field codes for the <oph:VerificationStatus> tag:
    00	Unknown	It is genuinely unknown whether the preservation agency has checked the preserved holdings in the <Coverage> statement. This status should be updated when information is available		
    01	Unverified	Preservation agency has not checked the preserved holdings in the <Coverage> statement, and it is not known whether all issues or e-content items are preserved
    02	Verification in progress	Preservation agency is in the process of checking the preserved holdings in the <Coverage> statement, and nothing can be said about completeness at this stage		
    03	Verified and incomplete	Preservation agency has checked the preserved holdings in the <Coverage> statement, and some issues or e-content items intended to be preserved are not complete		
    04	Verified and complete	Preservation agency has checked the preserved holdings in the <Coverage> statement, and all issues or e-content items intended to be preserved are complete
    """
    verified = models.IntegerField()

    publisher = models.ForeignKey(
        Publisher, on_delete=models.CASCADE, default=None
    )

    def __str__(self):
        return "{} (issn: {}) (v{}({})) ".format(
            self.title, self.issn, self.preserved_volumes, self.preserved_no
        )

    @staticmethod
    def name():
        return "OCUL Scholars Portal"

    @staticmethod
    def preservation(
        container_title, issn, volume, no=None, year=None, doi=None
    ):
        """
        Determine whether a DOI is preserved in OCUL Scholars Portal
        :param container_title: the container title
        :param issn: the ISSN
        :param volume: the volume
        :param no: the issue number
        :param year: the year
        :param doi: a DOI
        :return: An OculScholarsPortalPreservation item (or None) and a bool
        indicating whether the item is fully preserved
        """
        preserved_item = utils.get_preserved_item_record(
            OculScholarsPortalPreservation, container_title, issn
        )

        if not preserved_item or len(preserved_item) == 0:
            return None, False

        # direct hit
        pis = preserved_item.filter(preserved_volumes=volume)

        volume = str(volume)

        if len(pis) > 0:
            return pis[0], pis[0].preserved
        else:
            # do a broader search that allows for the following volume formats:
            # v24(1)
            # 1, 2, 3-4, 3.4
            # 1 to 4
            basic_volume_regex = r"v(\d+)"
            unpack_volume_regex = r"(\d+\-\d+)"
            unpack_volume_words_regex = r"(\d+ to \d+)"

            pis = preserved_item.filter(
                preserved_volumes__iregex=basic_volume_regex
            )

            if len(pis) > 0:
                for pi in pis:
                    matches = re.findall(
                        basic_volume_regex, pi.preserved_volumes
                    )

                    if volume in matches:
                        return pi, pi.preserved

            pis = preserved_item.filter(
                preserved_volumes__iregex=unpack_volume_regex
            )

            if len(pis) > 0:
                for pi in pis:
                    matches = re.findall(
                        unpack_volume_regex, pi.preserved_volumes
                    )

                    for match in matches:
                        unpacked_range = utils.unpack_range(match)

                        if volume in unpacked_range:
                            return pi, pi.preserved

            pis = preserved_item.filter(preserved_volumes=volume)
            return None, False

    @staticmethod
    def import_data(url, bucket="", s3client=None):
        # download the data file from S3 bucket
        with tempfile.TemporaryDirectory() as tmp:
            path = Path(tmp) / "downloaded.file"

            xml_file = utils.download_remote(
                False,
                OculScholarsPortalPreservation,
                url,
                bucket=bucket,
                s3client=s3client,
                decode=False,
                file=True,
                filename=str(path),
            )

            # local file for testing
            # xml_file = '/home/martin/scholars_portal_keepers_20230202.xml'

            # clear out
            utils.clear_out(OculScholarsPortalPreservation)

            fields = [
                "oph:ResourceVersionIDType",
                "oph:IDValue",
                "oph:FixedCoverage",
                "oph:PreservationStatusCode",
                "oph:VerificationStatus",
            ]

            xml_parsed = utils.process_onix(
                xml_file,
                fields,
                callback=OculScholarsPortalPreservation.create_preservation,
            )

    @staticmethod
    def create_preservation(output) -> None:
        """
        Create a preservation item of this model
        :param output: a dictionary from an ONIX import
        :return: None
        """

        if "status" not in output or output["status"] is None:
            status = 0
        else:
            status = output["status"]

        if "verified" not in output or output["verified"] is None:
            verified = 0
        else:
            verified = output["verified"]

        if "publisher" not in output or output["publisher"] is None:
            publisher = "Default"
        else:
            publisher = output["publisher"]

        if "title" not in output or output["title"] is None:
            title = "Unknown title"
        else:
            title = output["title"]

        publisher, created = Publisher.objects.get_or_create(name=publisher)

        volumes = output["volumes"]
        issues = output["issues"]

        for volume, issue in zip(volumes, issues):
            volume = volume if volume is not None else ""
            issue = issue if issue is not None else ""

            OculScholarsPortalPreservation.objects.create(
                issn=output["issn"],
                title=title,
                preserved_volumes=volume,
                preserved_no=issue,
                preserved=status,
                verified=verified,
                publisher=publisher,
            )

        logging.info(
            f"Added {title} to " f"{OculScholarsPortalPreservation.name()} data"
        )


class ISSNL(models.Model):
    class Meta:
        db_table = "preservationData_issnl"

        indexes = [
            models.Index(fields=["identifier"]),
        ]

    identifier = models.CharField(max_length=255)


class ISSN(models.Model):
    class Meta:
        db_table = "preservationData_issn"

        indexes = [
            models.Index(fields=["identifier"]),
        ]

    identifier = models.CharField(max_length=255)
    issnl = models.ForeignKey(ISSNL, on_delete=models.CASCADE, default=None)


class InternetArchivePreservation(models.Model):
    class Meta:
        db_table = "preservationData_iapreservation"

        indexes = [
            models.Index(fields=["issn"]),
            models.Index(fields=["eissn"]),
            models.Index(fields=["title"]),
        ]

    title = models.TextField()
    issn = models.CharField(max_length=20)
    eissn = models.CharField(max_length=20)
    preserved_volumes = models.TextField()

    # this indicates whether the title is preserved or queued
    preserved = models.BooleanField()
    publisher = models.ForeignKey(
        Publisher, on_delete=models.CASCADE, default=None
    )

    def __str__(self):
        return "{} (issn: {}) (eissn: {}) (v{}) ".format(
            self.title, self.issn, self.eissn, self.preserved_volumes
        )

    @staticmethod
    def name():
        return "Internet Archive"

    @staticmethod
    def preservation(
        container_title, issn, volume, no=None, year=None, doi=None
    ):
        """
        Determine whether a DOI is preserved in the Internet Archive
        :param container_title: the container title
        :param issn: the ISSN
        :param volume: the volume
        :param no: the issue number
        :param year: the year
        :param doi: a DOI
        :return: An InternetArchivePreservation item (or None) and a bool
        indicating whether the item is fully preserved
        """
        preserved_item = utils.get_preserved_item_record(
            InternetArchivePreservation, container_title, issn
        )

        if not preserved_item or len(preserved_item) == 0:
            return None, False

        # We add a single IA entry per volume
        pi = preserved_item.filter(preserved_volumes=str(volume))

        if pi.count() > 0:
            return pi[0], True

        return None, False

    @staticmethod
    @transaction.atomic
    def import_data(url, bucket="", s3client=None, local=False):
        # download the data file from S3 bucket
        with tempfile.TemporaryDirectory() as tmp:
            if local:
                path = url
            else:
                path = Path(tmp) / "downloaded.file"

                ia_data = utils.download_remote(
                    False,
                    InternetArchivePreservation,
                    url,
                    bucket=bucket,
                    s3client=s3client,
                    decode=False,
                    file=True,
                    filename=str(path),
                )

            # clear out existing data
            utils.clear_out(InternetArchivePreservation)

            csv.field_size_limit(sys.maxsize)

            with open(str(path), "r") as input_file:
                csv_reader = csv.DictReader(input_file, delimiter="\t")

                for row in csv_reader:
                    publisher, created = Publisher.objects.get_or_create(
                        name=row["publisher_name"]
                    )

                    # we have to calculate the held volumes
                    first_vol = row["num_first_vol_online"]
                    last_vol = row["num_last_vol_online"]

                    if first_vol != "":
                        try:
                            first_vol = int(first_vol)
                        except ValueError:
                            first_vol = None

                    if last_vol != "":
                        try:
                            last_vol = int(last_vol)
                        except ValueError:
                            last_vol = None

                    volumes = []

                    if first_vol and last_vol is None:
                        # if no last volume, assume single volume
                        volumes.append(first_vol)
                    elif last_vol and first_vol is None:
                        # if no first volume, assume single volume
                        volumes.append(last_vol)
                    elif first_vol and last_vol:
                        # if there's a first and last volume
                        volumes.extend(range(first_vol, last_vol))

                    volumes = set(volumes)

                    for vol in volumes:
                        InternetArchivePreservation.objects.create(
                            issn=row["print_identifier"],
                            eissn=row["online_identifier"],
                            title=row["publication_title"],
                            preserved_volumes=str(vol),
                            preserved=True,
                            publisher=publisher,
                        )

                    logging.info(
                        f'Added {row["publication_title"]} to '
                        f"{InternetArchivePreservation.name()} data "
                        f"({len(volumes)} volumes)"
                    )


class InternetArchiveItem(models.Model):
    class Meta:
        db_table = "preservationData_iaitems"

        indexes = [
            models.Index(fields=["doi"]),
            models.Index(fields=["fatcat_ident"]),
        ]

    doi = models.TextField()
    url = models.TextField()
    fatcat_ident = models.TextField()

    def __str__(self):
        return "{} (url: {}) (fatcat: {})".format(
            self.doi, self.url, self.fatcat_ident
        )

    @staticmethod
    def name():
        return "Internet Archive Scholar (item-level data)"

    @staticmethod
    def import_data(file: str) -> None:
        """Imports data from an Internet Archive file. Updates are propagated
        in batches of 1000.

        :param file: the file to import from
        :return None
        """
        from rich.progress import track
        from rich.progress import Progress

        from msgspec.json import decode
        from msgspec import Struct
        from typing import Optional

        class URL(Struct):
            url: str
            rel: str

        class Files(Struct):
            urls: list[URL]

        class Identifier(Struct):
            doi: Optional[str] = None

        class IAItem(Struct):
            ident: str
            ext_ids: Identifier
            files: list[Files]

        file_size = Path(file).stat().st_size

        with Progress() as progress:
            with open(file, "rb") as f:
                task = progress.add_task("[red]Importing...", total=file_size)

                count = 0
                sample_count = 0
                batch_objects = []

                utils.clear_out(InternetArchiveItem)

                for json_data in f:
                    data = decode(json_data, type=IAItem)

                    list_comp = [
                        (data, x)
                        for file in data.files
                        for x in file.urls
                        if x.rel == "webarchive"
                        and data.ext_ids.doi is not None
                    ]

                    progress.update(task, advance=len(json_data))

                    for outcome, url in list_comp:
                        batch_objects.append(
                            InternetArchiveItem(
                                doi=outcome.ext_ids.doi,
                                url=url,
                                fatcat_ident=outcome.ident,
                            )
                        )

                        count += 1

                    # this is just a logging function so we can see
                    # what's going on
                    if sample_count == 100000:
                        progress.console.print(
                            "Buffer contains {} items".format(count)
                        )
                        progress.console.print(f"{data}")
                        sample_count = 0

                    sample_count += 1

                    if count > 1000:
                        InternetArchiveItem.objects.bulk_create(batch_objects)
                        progress.console.print(
                            f"Flushing {len(batch_objects)} items."
                        )
                        count = 0
                        batch_objects = []

                # flush the final objects
                InternetArchiveItem.objects.bulk_create(batch_objects)
                progress.console.print(f"Flushing {len(batch_objects)} items.")

            return

    @staticmethod
    def preservation(
        container_title, issn, volume, no=None, year=None, doi=None
    ):
        """
        Determine whether a DOI is preserved in the Internet Archive
        :param container_title: the container title
        :param issn: the ISSN
        :param volume: the volume
        :param no: the issue number
        :param year: the year
        :param doi: a DOI
        :return: An InternetArchiveItem item (or None) and a bool
        indicating whether the item is fully preserved
        """
        try:
            preserved_item = InternetArchiveItem.objects.filter(doi=doi["DOI"])

            if len(preserved_item) > 0:
                return preserved_item[0], True
            else:
                return None, False
        except InternetArchiveItem.DoesNotExist:
            return None, False



import inspect
import logging
import os.path
import sys
import humanfriendly
import datetime

import boto3
import click
import pytz
from rich.console import Console
from rich.logging import RichHandler

FORMAT = "%(message)s"
logging.basicConfig(
    level=logging.INFO,
    format=FORMAT,
    datefmt="[%X]",
    handlers=[RichHandler(console=Console(stderr=True))],
)
log = logging.getLogger("preservation-database")
log.setLevel(logging.INFO)

sys.path.append(
    os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
)
# os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'settings')

import django

django.setup()

from preservationdatabase.models import (
    CarinianaPreservation,
    ClockssPreservation,
    HathiPreservation,
    InternetArchivePreservation,
    LockssPreservation,
    OculScholarsPortalPreservation,
    PKPPreservation,
    PorticoPreservation,
    LastFill,
    ISSN,
    ISSNL,
    InternetArchiveItem,
)

from django.db import transaction
import utils


@click.group()
def cli():
    pass


@click.command()
@click.option(
    "--url",
    default="https://api.portico.org/kbart/Portico_Holding_KBart.txt",
    help="The URL to fetch",
)
@click.option("--local", is_flag=True, default=False)
@transaction.atomic()
def import_portico(url, local):
    """Download and import data from Portico"""
    PorticoPreservation.import_data(url, local=local)


@click.command()
@click.option(
    "--file",
    default="ia_sim_keepers_kbart.2022-12-12.tsv",
    help="The filename of the Internet Archive full dump to use",
)
@click.option(
    "--bucket",
    default="preservation.research.crossref.org",
    help="The s3 bucket from which to retrieve the data",
)
@click.option("--local", is_flag=True, default=False)
@transaction.atomic()
def import_internet_archive(file, bucket, local):
    """Import data from Internet Archive (requires local file download or S3)"""
    s3client = boto3.client("s3")
    InternetArchivePreservation.import_data(
        file, bucket=bucket, s3client=s3client, local=local
    )


@click.command()
@click.option(
    "--file",
    default="release_export_expanded.json",
    help="The filename of the Internet Archive release_export_expanded to use",
)
def import_internet_archive_items(file):
    """Import item data from Internet Archive"""
    InternetArchiveItem.import_data(file)


@click.command()
@click.option(
    "--url",
    default="https://reports.clockss.org/keepers/keepers-CLOCKSS-report.csv",
    help="The URL to fetch",
)
@click.option("--local", is_flag=True, default=False)
@transaction.atomic()
def import_clockss(url, local):
    """Download and import data from CLOCKSS"""
    ClockssPreservation.import_data(url, local=local)


@click.command()
@click.option(
    "--url",
    default="https://reports.lockss.org/keepers/keepers-LOCKSS-report.csv",
    help="The URL to fetch",
)
@click.option("--local", is_flag=True, default=False)
@transaction.atomic()
def import_lockss(url, local):
    """Download and import data from LOCKSS"""
    LockssPreservation.import_data(url, local=local)


@click.command()
@click.option(
    "--url",
    default="https://pkp.sfu.ca/files/pkppn/onix.csv",
    help="The URL to fetch",
)
@click.option("--local", is_flag=True, default=False)
@transaction.atomic()
def import_pkp(url, local):
    """Download and import data from PKP's private LOCKSS network"""
    PKPPreservation.import_data(url, local=local)


@click.command()
@click.option(
    "--url",
    default="http://reports-lockss.ibict.br/keepers/pln/ibictpln/keepers-IBICTPLN-report.csv",
    help="The URL to fetch",
)
@click.option("--local", is_flag=True, default=False)
@transaction.atomic()
def import_cariniana(url, local):
    """Download and import data from Cariniana"""
    CarinianaPreservation.import_data(url, local=local)


@click.command()
@click.option(
    "--file",
    default="hathi_full_20230101.txt",
    help="The filename of the Hathitrust full dump to use",
)
@click.option(
    "--bucket",
    default="preservation.research.crossref.org",
    help="The s3 bucket from which to retrieve the data",
)
@click.option("--local", is_flag=True, default=False)
@transaction.atomic()
def import_hathi(file, bucket, local):
    """Import data from Hathi (requires local file download or S3)"""
    s3client = boto3.client("s3")
    HathiPreservation.import_data(
        file, bucket=bucket, s3client=s3client, local=local
    )


@click.command()
@click.option(
    "--file",
    default="scholars_portal_keepers_20230202.xml",
    help="The filename of the OCUL full dump to use",
)
@click.option(
    "--bucket",
    default="preservation.research.crossref.org",
    help="The s3 bucket from which to retrieve the data",
)
@transaction.atomic()
def import_ocul(file, bucket):
    """Import data from Ocul (requires local file download or S3)"""

    s3client = boto3.client("s3")
    OculScholarsPortalPreservation.import_data(
        file, bucket=bucket, s3client=s3client
    )

    return


@click.command()
@transaction.atomic()
def stamp_cache_today():
    """Mark the latest imports as today"""
    from preservationdatabase.constants import archives

    for key, value in archives.items():
        LastFill.set_last_fill(value.name())


@click.command()
@transaction.atomic()
def clear_cache():
    """Clear the import cache"""
    LastFill.clear()


@click.command()
@click.argument("issnl_file")
@transaction.atomic()
def import_issnl(issnl_file):
    """Import ISSN-L mappings"""
    utils.import_issnl(issnl_file, ISSNL, ISSN)


@click.command()
@transaction.atomic()
def show_cache():
    """Show last fill date/times and cache status"""

    for lf in LastFill.objects.all():
        logging.info(lf)
        time_delta = datetime.datetime.now(pytz.utc) - lf.last_fill_date
        logging.info(
            "Last cache stamp for {} was {}".format(
                lf.archive_name,
                humanfriendly.format_timespan(
                    humanfriendly.coerce_seconds(time_delta)
                ),
            )
        )

        if lf.cache_valid:
            logging.info("{} will use cached version".format(lf.archive_name))
        else:
            logging.info(
                "{} will be fetched from source".format(lf.archive_name)
            )


@click.command()
@transaction.atomic()
def show_archives():
    """Clear the import cache"""
    from constants import archives

    for key, value in archives.items():
        logging.info(value.name())


@click.command()
@click.option("--resume-from", default=0, help="Resume from this member ID")
@transaction.atomic()
def process_members(resume_from: int):
    """Process each member"""
    import exporter

    if resume_from > 0:
        logging.info("Resuming from member ID {}".format(resume_from))

    exporter.generate_report(resume_from)


@click.command()
@click.option(
    "--limit", default=0, help="Run a report on a partial subset of data"
)
@transaction.atomic()
def overall_report(limit: int = 0):
    """Process each member"""
    import exporter

    if limit > 0:
        logging.info("Running a limited report on {} members".format(limit))

    import json

    print(json.dumps(exporter.overall_report(members=None, limit=limit)))


@click.command()
@click.argument("member_id")
@transaction.atomic()
def member_report(member_id):
    """Process the member"""

    import exporter

    exporter.member_report(member_id)


@click.command()
@transaction.atomic()
def member_reports():
    """Process all members"""

    import exporter

    exporter.member_reports()


@click.command()
@click.argument("member_id")
@transaction.atomic()
def process_member(member_id):
    """Process a member"""
    SAMPLES_BUCKET = "samples.research.crossref.org"
    SAMPLES_PATH = "members-works/2023-03-05/samples/"

    ANNOTATION_BUCKET = "outputs.research.crossref.org"
    ANNOTATION_PATH = "annotations"
    ANNOTATION_FILENAME = "preservation.json"

    CODE_BUCKET = "airflow-crossref-research-annotation"

    PARALLEL_JOBS = 5

    import exporter

    exporter.process_sample(
        ANNOTATION_BUCKET,
        ANNOTATION_FILENAME,
        ANNOTATION_PATH,
        SAMPLES_BUCKET,
        SAMPLES_PATH,
        member_id,
        CODE_BUCKET,
        verbose=False,
    )


@click.command()
@click.argument("issn")
@transaction.atomic()
def show_issn(issn):
    """Show preservation items that match an ISSN"""
    from constants import archives

    for key, archive_object in archives.items():
        issns = archive_object.objects.filter(issn=issn)

        for issn_object in issns:
            logging.info(f"{archive_object.name()} preserves {issn_object}")

        if "eissn" in archive_object._meta.get_fields():
            issns = archive_object.objects.filter(eissn=issn)

            for issn_object in issns:
                logging.info(f"{archive_object.name()} preserves {issn_object}")


@click.command()
@click.argument("archive_name")
@click.option("--count", default=25, help="Give random samples for an archive")
@click.option(
    "--only-crossref",
    help="Only return Crossref titles",
    is_flag=True,
    default=False,
)
@transaction.atomic()
def random_samples(archive_name, count, only_crossref):
    """Return random samples that occur in and out of an archive"""
    from constants import archives
    from crossref.restful import Etiquette, Journals
    import utils

    archive = None

    for key, archive_object in archives.items():
        if archive_object.name() == archive_name:
            archive = archive_object

    if not archive:
        log.error(f"Archive {archive_name} not found")
        return

    random_entries = utils.random_db_entries(
        archive, count, only_crossref=only_crossref
    )
    not_in_archive = []

    my_etiquette = Etiquette(
        "Preservation Status", "0.01", "https://eve.gd", "meve@crossref.org"
    )

    jrnls = Journals(etiquette=my_etiquette)

    # now find entries that are not in the archive:
    while len(not_in_archive) < count - 1:
        for key, other_archive in archives.items():
            if other_archive != archive:
                other_entries = utils.random_db_entries(other_archive, count)

                for entry in other_entries:
                    if not utils.in_archive(entry, archive):
                        if only_crossref:
                            append = False

                            new_issns = utils.expand_issns([entry.issn])

                            for issn in new_issns:
                                jrnl = jrnls.journal(issn=issn)

                                if jrnl:
                                    append = True
                                    break

                            if hasattr(entry, "eissn"):
                                new_eissns = utils.expand_issns([entry.eissn])

                                for issn in new_eissns:
                                    jrnl = jrnls.journal(issn=issn)

                                    if jrnl:
                                        append = True
                                        break

                            if append:
                                not_in_archive.append(entry)
                                print(
                                    "Found not-in-archive crossref entry ({}): {}".format(
                                        len(not_in_archive), entry
                                    )
                                )
                        else:
                            not_in_archive.append(entry)

                        if len(not_in_archive) == count - 1:
                            break

            if len(not_in_archive) == count - 1:
                break

    for entry in random_entries:
        logging.info("{} contains: {}".format(archive_name, entry))

    for entry in not_in_archive:
        logging.info("{} does not contain: {}".format(archive_name, entry))


@click.command()
def draw_graphs():
    import exporter

    # exporter.draw_graphs()
    pass


@click.command()
def import_all():
    """Download and import all data (excluding HathiTrust)"""

    import_clockss(
        url="https://reports.clockss.org/keepers/keepers-CLOCKSS-report.csv"
    )

    import_portico(
        url="https://api.portico.org/kbart/Portico_Holding_KBart.txt"
    )

    import_lockss(
        url="https://reports.lockss.org/keepers/keepers-LOCKSS-report.csv"
    )

    import_cariniana(
        url="http://reports-lockss.ibict.br/keepers/pln/ibictpln/keepers-IBICTPLN-report.csv"
    )


@click.command()
def rescore():
    """Rescore all data"""
    import exporter

    exporter.rescore_all()


@click.command()
@click.argument("member_id")
def score_member(member_id):
    """Score a member and write to console only"""
    import exporter

    exporter.score_member(member_id)


@click.command()
@click.argument("member_id")
def member_position(member_id):
    """Show how far into the members list a member is"""
    import exporter

    exporter.member_position(member_id=member_id)


@click.command()
def rename():
    """Rename keys to kebab case"""
    import exporter

    exporter.rename_keys()


@click.command()
@click.argument("doi")
@click.option(
    "--literal", help="Use literal DOI (no schema)", is_flag=True, default=False
)
@click.option(
    "--no-issn", help="Do not expand ISSN list", is_flag=True, default=False
)
def show_preservation(doi, literal=False, no_issn=False):
    """
    Determine whether a DOI is preserved
    """
    if not literal:
        doi = utils.normalize_doi(doi)

    preservation_statuses, doi = utils.show_preservation_for_doi(
        doi, no_issn=no_issn
    )

    for key, value in preservation_statuses.items():
        preserved, done = value

        if preserved:
            if done:
                log.info(
                    f"[green]Preserved:[/] in {key}", extra={"markup": True}
                )
            else:
                log.info(
                    f"[yellow]Preserved (in progress):[/] " f"in {key}",
                    extra={"markup": True},
                )
        else:
            log.info(f"[red]Not preserved:[/] in {key}", extra={"markup": True})


def test():
    print("RUNNING")


if __name__ == "__main__":
    cli.add_command(clear_cache)
    cli.add_command(draw_graphs)
    cli.add_command(import_all)
    cli.add_command(import_cariniana)
    cli.add_command(import_clockss)
    cli.add_command(import_hathi)
    cli.add_command(import_internet_archive)
    cli.add_command(import_internet_archive_items)
    cli.add_command(import_issnl)
    cli.add_command(import_lockss)
    cli.add_command(import_ocul)
    cli.add_command(import_pkp)
    cli.add_command(import_portico)
    cli.add_command(member_position)
    cli.add_command(member_report)
    cli.add_command(member_reports)
    cli.add_command(overall_report)
    cli.add_command(process_member)
    cli.add_command(process_members)
    cli.add_command(random_samples)
    cli.add_command(rename)
    cli.add_command(rescore)
    cli.add_command(score_member)
    cli.add_command(show_archives)
    cli.add_command(show_cache)
    cli.add_command(show_issn)
    cli.add_command(show_preservation)
    cli.add_command(stamp_cache_today)
    cli()



import hashlib
import json
import logging
import os
import sys
from collections import defaultdict, OrderedDict
from datetime import date, datetime
from typing import DefaultDict, Any

from caseconverter import kebabcase

SAMPLES_BUCKET = "samples.research.crossref.org"
SAMPLES_PATH = "members-works/2023-03-05/samples/"

ANNOTATION_BUCKET = "outputs.research.crossref.org"
ANNOTATION_PATH = "annotations"
ANNOTATION_FILENAME = "preservation.json"

REPORT_BUCKET = "outputs.research.crossref.org"
REPORT_PATH = "reports"
REPORT_FILENAME = "preservation.json"

CODE_BUCKET = "airflow-crossref-research-annotation"

PARALLEL_JOBS = 5

import boto3


def get_members(
    s3client=None, samples_bucket=None, samples_path=None
) -> list[str]:
    """
    Retrieves the list of members from the S3 bucket
    :param s3client: the s3client object to use to fetch
    :param samples_bucket: the name of the samples bucket
    :param samples_path: the path of the samples object
    :return: a list of strings of member IDs
    """

    # the samples research bucket contains JSON-L with the filename
    # schema member-1.jsonl etc.
    # Note also: we use S3 for this so that we definitely know what
    # files are available, rather than the REST API (in case there is
    # a bug in the sampling framework that leads to an unavailable
    # object)

    if not s3client:
        s3client = boto3.client("s3")

    import re

    r = re.compile(r"member-(\d+).jsonl")
    return [
        m.group(1)
        for m in map(
            r.match,
            list_bucket(
                s3client,
                samples_bucket=samples_bucket,
                samples_path=samples_path,
            ),
        )
        if m is not None
    ]


def list_bucket(s3client, samples_bucket, samples_path) -> list[str]:
    """
    Lists the contents of the samples bucket
    :param s3client: the s3client object to use to fetch
    :param samples_bucket: the name of the samples bucket
    :param samples_path: the path of the samples object
    :return: a list of object names
    """

    paginator = s3client.get_paginator("list_objects_v2")
    member_list = []

    for page in paginator.paginate(
        Bucket=samples_bucket, Prefix=f"{samples_path}"
    ):
        for obj in page["Contents"]:
            filename = obj["Key"]
            member_list.append(filename.split("/")[-1])

    return member_list


def get_samples(
    s3client, member_id, samples_bucket, samples_path, verbose=False
) -> list:
    """
    Retrieves the list of samples from the S3 bucket
    :param s3client: the s3client object to use to fetch
    :param samples_bucket: the name of the samples bucket
    :param samples_path: the path of the samples object
    :param member_id: the ID of the member to retrieve
    :param verbose: whether to log progress
    :return: a list of samples
    """

    key = f"{samples_path}member-{member_id}.jsonl"

    data = (
        s3client.get_object(Bucket=samples_bucket, Key=key)["Body"]
        .read()
        .decode("utf-8")
    )

    from io import StringIO
    import logging

    with StringIO(data) as json_file:
        output = list(json_file)
        if verbose:
            logging.info(
                f"Found {len(output)} samples for " f"member {member_id}"
            )
        return output


def preservation_status(result) -> (dict, str):
    """
    Return preservation statistics for a specific member
    :param result: the pre-parsed JSON entry of the member
    :return: 2-tuple: dictionary of preservations and the DOI string
    """
    from utils import show_preservation

    container_title = (
        result["container-title"] if "container-title" in result else None
    )
    issn = result["ISSN"] if "ISSN" in result else None
    volume = result["volume"] if "volume" in result else None

    # not in sampling framework (yet)
    no = None

    year = None
    # "published": {"date-parts": [[2005, 12, 30]]}}}

    if "published" in result:
        year = result["published"]["date-parts"][0][0]

    return show_preservation(
        container_title=container_title,
        issn=issn,
        volume=volume,
        no=no,
        doi=result,
        year=year,
    )


def score(
    one_archive, two_archives, three_archives, sample_count, verbose=False
):
    score = "Unclassified"

    if sample_count > 0:
        if verbose:
            logging.info(
                "Score for Bronze is: {}. Needs to be greater than 0.25".format(
                    (one_archive / sample_count)
                    + (two_archives / sample_count)
                    + (three_archives / sample_count)
                )
            )

            logging.info(
                "Score for Silver is: {}. Needs to be greater than 0.50".format(
                    (two_archives / sample_count)
                    + (three_archives / sample_count)
                )
            )

            logging.info(
                "Score for Gold is: {}. Needs to be greater than 0.75".format(
                    (three_archives / sample_count)
                )
            )

        # 25% in 1 or more archives
        if (
            (one_archive / sample_count)
            + (two_archives / sample_count)
            + (three_archives / sample_count)
        ) >= 0.25:
            score = "Bronze"

        # 50% in 2 or more archives
        if (
            (two_archives / sample_count) + (three_archives / sample_count)
        ) >= 0.50:
            score = "Silver"

        # 75% in 3 archives
        if (three_archives / sample_count) >= 0.75:
            score = "Gold"

    return score


def process_member_sample(samples, sample_path, verbose=False) -> dict:
    """
    Processes samples for a single member
    :param samples: the samples to process
    :param sample_path: the path of the sample
    :param verbose: whether to print the sample
    :return: a dictionary of preservation statistics
    """
    from constants import archives
    from datetime import datetime
    import json

    from collections import defaultdict

    overall_status: DefaultDict[Any, Any] = defaultdict(int)

    # date stamp this output
    overall_status["about"] = {
        "date-generated": str(datetime.now()),
        "sample-file": sample_path,
    }

    three_archives = 0
    two_archives = 0
    one_archive = 0

    session = boto3.Session()
    s3 = session.resource("s3")

    for sample_item in samples:
        result = json.loads(sample_item)["data-point"]

        overall_status["about"]["member"] = result["member"]
        overall_status["unpreserved-items"] = list()

        # we can only work with journal articles
        # we exclude journal articles from the current year because they
        # most likely have not been ingested into digital preservation
        # systems yet
        if "type" in result and result["type"] == "journal-article":
            year_published = (
                result["published"]["date-parts"][0][0]
                if "published" in result
                else None
            )

            if year_published and str(year_published) != str(date.today().year):
                has_preservation = False
                archive_count = 0

                preservation_statuses, doi = preservation_status(result)

                md5_doi = hashlib.md5(doi["DOI"].lower().encode()).hexdigest()
                aws_key = (
                    f"{ANNOTATION_PATH}/works/{md5_doi}/{ANNOTATION_FILENAME}"
                )

                preserved_json = {
                    "about": {
                        "DOI": doi["DOI"],
                        "date-generated": str(datetime.now()),
                    }
                }

                # increment the sample count
                overall_status["sample-count"] += 1

                for key, value in preservation_statuses.items():
                    preserved, done = value

                    if preserved:
                        has_preservation = True

                        # increment this archive's stats
                        overall_status[key] += 1

                        # increment the archive counter
                        archive_count += 1

                        # increment total preservation instances count
                        overall_status["preservation-instances"] += 1

                        # log this for DOI-level annotation
                        if done:
                            preserved_json[key] = "preserved"
                        else:
                            preserved_json[key] = "preserved (in progress)"

                    else:
                        preserved_json[key] = "unpreserved"

                new_annotation = {}

                for key, value in preserved_json.items():
                    new_key = normalize_json_key(key)
                    new_annotation[new_key] = value

                push_json_to_s3(
                    s3=s3,
                    json_obj=new_annotation,
                    bucket=ANNOTATION_BUCKET,
                    path=aws_key,
                )

                # preserved_count refers to the number of works with at
                # least one preservation
                if has_preservation:
                    overall_status["preserved-count"] += 1

                    if verbose:
                        logging.info(f"{doi['DOI']} has a preservation")
                        logging.info(
                            f"Running total: "
                            f"{overall_status['preserved-count']} preserved, "
                            f"{overall_status['unpreserved-count']} unpreserved"
                        )

                else:
                    overall_status["unpreserved-count"] += 1

                    overall_status["unpreserved-items"].append(doi["DOI"])

                    if verbose:
                        logging.info(f"{doi['DOI']} has no preservation")
                        logging.info(
                            f"Running total: "
                            f"{overall_status['preserved-count']} preserved, "
                            f"{overall_status['unpreserved-count']} unpreserved"
                        )

                # increment the correct counters for overall stats
                if archive_count == 1:
                    one_archive += 1
                elif archive_count == 2:
                    two_archives += 1
                elif archive_count > 2:
                    three_archives += 1
            else:
                overall_status["excluded-current-year"] += 1
                if verbose:
                    logging.info(
                        f"{result['DOI']} is excluded as is too recent"
                    )
        else:
            # this is an excluded sample item
            if "type" in result and result["type"] != "journal-article":
                overall_status["excluded-non-journal"] += 1
                if verbose:
                    logging.info(
                        f"{result['DOI']} is excluded as it is not a journal "
                        f"article"
                    )
            elif "published" not in result:
                overall_status["excluded-no-date"] += 1
                if verbose:
                    logging.info(
                        f"{result['DOI']} is excluded as it is not a journal "
                        f"article"
                    )
            elif date.today().year == result["published"]["date-parts"][0][0]:
                overall_status["excluded-current-year"] += 1
                if verbose:
                    logging.info(
                        f"{result['DOI']} is excluded as is too recent"
                    )

    # add blank keys for archives that weren't used
    for preservation_system, class_o in archives.items():
        if class_o.name() not in overall_status:
            overall_status[class_o.name()] = 0

    # calculate percentage of preservation
    if overall_status["sample-count"] > 0:
        overall_status["percentage-preserved"] = (
            overall_status["preserved-count"] / overall_status["sample-count"]
        ) * 100
    else:
        overall_status["percentage-preserved"] = 0

    # add the counters
    overall_status["preserved-in-one-archive"] = one_archive
    overall_status["preserved-in-two-archives"] = two_archives
    overall_status["preserved-in-three-or-more-archives"] = three_archives

    # determine the classes
    overall_status["member-grade"] = score(
        one_archive,
        two_archives,
        three_archives,
        overall_status["sample-count"],
    )

    return overall_status


def score_member(member_id):
    return rescore_member(
        ANNOTATION_BUCKET,
        ANNOTATION_FILENAME,
        ANNOTATION_PATH,
        SAMPLES_BUCKET,
        SAMPLES_PATH,
        member_id,
        CODE_BUCKET,
        push_to_s3=False,
        verbose=True,
    )


def push_json_to_s3(
    s3,
    json_obj,
    bucket,
    path,
    verbose=False,
) -> None:
    """
    Writes the JSON data to S3
    :param s3: the s3 object to use
    :param bucket: the name of the bucket
    :param path: the path of the object to store
    :param json_obj: the JSON to write
    :param verbose: whether to print status messages
    :return:
    """
    import json
    import logging

    if verbose:
        logging.info("Normalizing JSON keys")

    json_obj = normalize_json_keys(json_obj)

    if verbose:
        logging.info(f"Writing JSON to S3 bucket {bucket} at path {path}")

    obj = s3.Object(bucket, path)
    obj.put(Body=json.dumps(json_obj))


def normalize_json_keys(json_obj: dict) -> dict:
    """
    Always normalize keys to kebab case
    :param json_obj: the JSON object to normalize
    :return: a new JSON object with normalized keys
    """
    new_json_obj = {}

    if isinstance(json_obj, list):
        new_json_obj = []

        for item in json_obj:
            new_item = {}

            for key in item:
                new_item[normalize_json_key(key)] = item[key]

            new_json_obj.append(new_item)
    else:
        for key in json_obj.keys():
            new_json_obj[normalize_json_key(key)] = json_obj[key]

    return new_json_obj


def normalize_json_key(key: str) -> str:
    """
    Normalize keys to kebab case
    :param key: the key to normalize
    :return: the normalized key
    """
    # while it seems overkill to have this as a separate function,
    # it's useful in other contexts, too, to have it on its own
    return kebabcase(key)


def process_sample(
    annotation_bucket,
    annotation_filename,
    annotation_path,
    samples_bucket,
    samples_path,
    member_id,
    code_bucket,
    verbose=False,
):
    """
    Process a single member sample
    :param annotation_bucket: the annotation bucket
    :param annotation_filename: the annotation filename
    :param annotation_path: the annotation path
    :param samples_bucket: the samples bucket
    :param samples_path: the samples path
    :param member_id: the member id
    :param verbose: whether to print status messages
    :param code_bucket: the code bucket where settings are located
    :return:
    """
    import logging

    # the environment setup is needed for logging etc. because of the
    # joblib parallelization
    import environment

    environment.setup_environment(code_bucket, download_settings=False)

    import boto3

    s3client = boto3.client("s3")
    session = boto3.Session()
    s3 = session.resource("s3")

    if verbose:
        logging.info(f"Processing member {member_id}")

    samples = get_samples(
        s3client,
        member_id,
        samples_bucket=samples_bucket,
        samples_path=samples_path,
    )

    overall_status = process_member_sample(
        samples, samples_path, verbose=verbose
    )

    key = f"{annotation_path}/members/{member_id}/{annotation_filename}"

    push_json_to_s3(
        s3=s3,
        json_obj=overall_status,
        bucket=annotation_bucket,
        path=key,
        verbose=True,
    )

    return {member_id: overall_status}


def generate_report(resume_from: int):
    import boto3
    import logging
    from rich.progress import track

    s3client = boto3.client("s3")

    from joblib import Parallel, delayed

    # get member list from S3
    members = get_members(
        s3client=s3client,
        samples_bucket=SAMPLES_BUCKET,
        samples_path=SAMPLES_PATH,
    )

    logging.info(f"There are {len(members)} to process.")

    # we resume from the resume form, in the order returned by the S3 bucket
    if resume_from > 0:
        index_slice = members.index(str(resume_from))
        members = members[index_slice:]

        logging.info(f"After resume, there are now {len(members)} to process.")

    # the results of the parallel processing are dictionaries with
    # the member ID as the key and a dictionary of preservation statistics
    # as the value.

    results = Parallel(n_jobs=PARALLEL_JOBS)(
        delayed(process_sample)(
            ANNOTATION_BUCKET,
            ANNOTATION_FILENAME,
            ANNOTATION_PATH,
            SAMPLES_BUCKET,
            SAMPLES_PATH,
            member_id,
            CODE_BUCKET,
        )
        for member_id in track(members)
    )

    return members


def member_reports():
    s3client = boto3.client("s3")
    session = boto3.Session()
    s3 = session.resource("s3")

    # get member list from S3
    members = get_members(
        s3client=s3client,
        samples_bucket=SAMPLES_BUCKET,
        samples_path=SAMPLES_PATH,
    )

    from joblib import Parallel, delayed
    from rich.progress import track

    results = Parallel(n_jobs=PARALLEL_JOBS)(
        delayed(member_report)(
            member_id,
        )
        for member_id in track(members)
    )


def member_report(member_id):
    import boto3
    import logging

    s3client = boto3.client("s3")
    session = boto3.Session()
    s3 = session.resource("s3")

    samples = get_samples(
        s3client=s3client,
        member_id=member_id,
        samples_bucket=SAMPLES_BUCKET,
        samples_path=SAMPLES_PATH,
    )

    overall_samples = []
    unpreserved_samples = []

    import botocore

    for sample in samples:
        sample = json.loads(sample)["data-point"]

        if "type" in sample and sample["type"] == "journal-article":
            year = None

            if "published" in sample:
                year = sample["published"]["date-parts"][0][0]

            if sample["DOI"] == "10.16995/glossa.8784":
                print("FOUND ARTICLE")
                print(str(year))
                print(str(datetime.now().year))
                print(str(year) != str(datetime.now().year))
                print(sample)

            if str(year) != str(datetime.now().year):
                md5_doi = hashlib.md5(
                    sample["DOI"].lower().encode()
                ).hexdigest()

                try:
                    annotation = get_annotation(
                        s3client=s3client,
                        member_id=member_id,
                        annotation_name="preservation.json",
                        annotations_bucket=ANNOTATION_BUCKET,
                        annotations_path=ANNOTATION_PATH,
                        path=f"works/{md5_doi}",
                        use_member_id=False,
                    )

                    preserved = any(
                        v.startswith("preserved")
                        for v in annotation.values()
                        if isinstance(v, str)
                    )

                    if not preserved:
                        unpreserved_samples.append(annotation)

                    overall_samples.append(annotation)

                except Exception as e:
                    ...
                    # print(e)
                    # print(f"This DOI was excluded: {sample['DOI']}")

    key = f"{REPORT_PATH}/members/{member_id}/preservation-report.json"
    unpreserved_key = f"{REPORT_PATH}/members/{member_id}/unpreserved.json"

    push_json_to_s3(
        s3=s3,
        json_obj=unpreserved_samples,
        bucket=REPORT_BUCKET,
        path=unpreserved_key,
        verbose=True,
    )

    push_json_to_s3(
        s3=s3,
        json_obj=overall_samples,
        bucket=REPORT_BUCKET,
        path=key,
        verbose=True,
    )


def rescore_member(
    annotation_bucket,
    annotation_filename,
    annotation_path,
    samples_bucket,
    samples_path,
    member_id,
    code_bucket,
    verbose=False,
    push_to_s3=True,
):
    import environment

    environment.setup_environment(code_bucket, download_settings=False)

    import boto3
    import logging

    s3client = boto3.client("s3")
    session = boto3.Session()
    s3 = session.resource("s3")

    try:
        preservations = get_annotation(
            s3client=s3client,
            annotations_bucket=ANNOTATION_BUCKET,
            annotations_path=ANNOTATION_PATH,
            member_id=member_id,
            annotation_name="preservation.json",
        )

        one_archive = preservations["preserved-in-one-archive"]
        two_archives = preservations["preserved-in-two-archives"]
        three_archives = preservations["preserved-in-three-or-more-archives"]

        sample_count = preservations["sample-count"]

        # determine the classes
        preservations["member-grade"] = score(
            one_archive,
            two_archives,
            three_archives,
            preservations["sample-count"],
            verbose=verbose,
        )

        key = f"{annotation_path}/members/{member_id}/{annotation_filename}"

        if verbose:
            logging.info(
                f"Rescoring member {member_id} as "
                f"{preservations['member-grade']}"
            )

        if push_to_s3:
            push_json_to_s3(
                s3=s3,
                json_obj=preservations,
                bucket=annotation_bucket,
                path=key,
            )

        return preservations["member-grade"]
    except Exception as e:
        logging.error(f"Unable to rescore member {member_id}. Error follows.")
        raise e


def overall_report(members=None, limit=0):
    import boto3
    import logging

    s3client = boto3.client("s3")
    session = boto3.Session()
    s3 = session.resource("s3")

    if not members:
        # get member list from S3
        members = get_members(
            s3client=s3client,
            samples_bucket=SAMPLES_BUCKET,
            samples_path=SAMPLES_PATH,
        )

    if limit > 0:
        members = members[:limit]

    # overall reports we want to build:
    # 1. Breakdown by member size
    # 2. Totally unpreserved members ["percentage-preserved"]
    # 3. Members with 75% in three archives (gold standard) ["member-grade"]
    # 4. Members with 50% in two archives (silver standard) ["member-grade"]
    # 5. Members with 25% in one archive (bronze standard) ["member-grade"]

    member_tier_names = [
        "<USD 1 million",
        "USD 1 million - USD 5 million",
        "USD 5 million - USD 10 million",
        "USD 10 million - USD 25 million",
        "USD 25 million - USD 50 million",
        "USD 50 million - USD 100 million",
        "USD 100 million - USD 200 million",
        "USD 200 million - USD 500 million",
        ">USD 500 million",
    ]

    member_tiers = {
        "<USD 1 million": 275,
        "USD 1 million - USD 5 million": 550,
        "USD 5 million - USD 10 million": 1650,
        "USD 10 million - USD 25 million": 3900,
        "USD 25 million - USD 50 million": 8300,
        "USD 50 million - USD 100 million": 14000,
        "USD 100 million - USD 200 million": 22000,
        "USD 200 million - USD 500 million": 33000,
        ">USD 500 million": 50000,
    }

    size_bracket_names = [
        "Unknown",
        "1-10",
        "11-100",
        "101-500",
        "501-1,000",
        "1,001-10,000",
        "10,001-100,000",
        "100,001-1,000,000",
        "1,000,001+",
    ]

    size_brackets = OrderedDict(
        {
            0: 0,
            1: 10,
            11: 100,
            101: 500,
            501: 1000,
            1001: 10000,
            10001: 100000,
            100001: 1000000,
            1000001: -1,
        }
    )

    gold_members = {member_tier: 0 for member_tier in member_tier_names}
    silver_members = {member_tier: 0 for member_tier in member_tier_names}
    bronze_members = {member_tier: 0 for member_tier in member_tier_names}
    unclassified_members = {member_tier: 0 for member_tier in member_tier_names}

    gold_members_by_country = defaultdict(int)
    silver_members_by_country = defaultdict(int)
    bronze_members_by_country = defaultdict(int)
    unclassified_members_by_country = defaultdict(int)

    gold_members_by_size = {
        size_bracket: 0 for size_bracket in size_bracket_names
    }
    silver_members_by_size = {
        size_bracket: 0 for size_bracket in size_bracket_names
    }
    bronze_members_by_size = {
        size_bracket: 0 for size_bracket in size_bracket_names
    }
    unclassified_members_by_size = {
        size_bracket: 0 for size_bracket in size_bracket_names
    }

    gold_member_ids = []
    silver_member_ids = []
    bronze_member_ids = []
    unclassified_member_ids = []

    total_items_unpreserved = 0
    total_items_preserved = 0
    total_preservation_instances = 0
    total_items_sampled = 0

    plus_token = os.environ.get("CROSSREF_PLUS_TOKEN", None)

    from crossref.restful import Etiquette, Members

    my_etiquette = Etiquette(
        "Preservation Status",
        "0.01",
        "https://eve.gd",
        "meve@crossref.org",
    )

    for member in members:
        logging.info(f"Fetching preservation statistics for member {member}")

        if plus_token:
            members = Members(
                etiquette=my_etiquette,
                crossref_plus_token=plus_token,
            )
            logging.info(f"Using plus lookup")
        else:
            members = Members(etiquette=my_etiquette)
            logging.info(f"Using regular lookup")

        logging.info(f"API lookup for {member}: done")

        member_obj = members.member(member)

        if "primary-name" not in member_obj:
            print(member_obj)

        preservations = get_annotation(
            s3client=s3client,
            annotations_bucket=ANNOTATION_BUCKET,
            annotations_path=ANNOTATION_PATH,
            member_id=member,
            annotation_name="preservation.json",
        )

        member_data = get_annotation(
            s3client=s3client,
            member_id=member,
            annotations_bucket=ANNOTATION_BUCKET,
            annotations_path=ANNOTATION_PATH,
            annotation_name="member-profile.json",
        )

        logging.info(f"S3 fetches for member {member}: done")

        try:
            annual_fee = int(member_data["annual-fee"])
        except ValueError:
            annual_fee = 0

        member_band = None

        logging.info(
            f"{member_obj['primary-name']} has an annual fee of {annual_fee}"
        )

        for key, val in member_tiers.items():
            if annual_fee <= int(val):
                # put this member in this band
                logging.info(f"{member_obj['primary-name']} is in band {key}")
                member_band = key
                break

            if member_band:
                break

        if not member_band:
            logging.warning(f"Unable to classify this member ({member})")

        # now log the country here
        country = member_obj["location"].split(",")[-1]

        # size em up
        deposits = member_obj["counts"]["total-dois"]

        size = None

        for key, val in size_brackets.items():
            if deposits == 0:
                size = "Unknown"
            elif val == -1:
                # special case
                if deposits >= key:
                    size = size_bracket_names[
                        list(size_brackets.keys()).index(key)
                    ]
            elif key <= deposits <= val:
                size = size_bracket_names[list(size_brackets.keys()).index(key)]

        total_items_preserved += (
            preservations["preserved-count"]
            if "preserved-count" in preservations
            else 0
        )
        total_items_unpreserved += (
            preservations["unpreserved-count"]
            if "unpreserved-count" in preservations
            else 0
        )
        total_preservation_instances += (
            preservations["preservation-instances"]
            if "preservation-instances" in preservations
            else 0
        )
        total_items_sampled += (
            preservations["sample-count"]
            if "sample-count" in preservations
            else 0
        )

        # this calculates the preservation grades for different member tiers
        # using fee level as the benchmark
        if preservations["member-grade"] == "Bronze":
            logging.info(f"{member_obj['primary-name']} is Bronze")
            bronze_members[member_band] += 1
            bronze_members_by_country[country] += 1
            bronze_members_by_size[size] += 1
            bronze_member_ids.append(f"{member_obj['primary-name']} ({member})")
        elif preservations["member-grade"] == "Silver":
            logging.info(f"{member_obj['primary-name']} is Silver")
            silver_members[member_band] += 1
            silver_members_by_country[country] += 1
            silver_members_by_size[size] += 1
            silver_member_ids.append(f"{member_obj['primary-name']} ({member})")
        elif preservations["member-grade"] == "Gold":
            logging.info(f"{member_obj['primary-name']} is Gold")
            gold_members[member_band] += 1
            gold_members_by_country[country] += 1
            gold_members_by_size[size] += 1
            gold_member_ids.append(f"{member_obj['primary-name']} ({member})")
        else:
            logging.info(f"{member_obj['primary-name']} is Unclassified")
            unclassified_members[member_band] += 1
            unclassified_members_by_country[country] += 1
            unclassified_members_by_size[size] += 1
            unclassified_member_ids.append(
                f"{member_obj['primary-name']} ({member})"
            )

    gold_totals = sum(gold_members.values())
    silver_totals = sum(silver_members.values())
    bronze_totals = sum(bronze_members.values())
    unclassified_totals = sum(unclassified_members.values())

    # bundle this all up
    overall_result = {
        "gold-members-by-fee": gold_members,
        "silver-members-by-fee": silver_members,
        "bronze-members-by-fee": bronze_members,
        "unclassified-members-by-fee": unclassified_members,
        "gold-totals": gold_totals,
        "silver-totals": silver_totals,
        "bronze-totals": bronze_totals,
        "unclassified-totals": unclassified_totals,
        "gold-members-by-country": gold_members_by_country,
        "silver-members-by-country": silver_members_by_country,
        "bronze-members-by-country": bronze_members_by_country,
        "unclassified-members-by-country": unclassified_members_by_country,
        "gold-members-by-size": gold_members_by_size,
        "silver-members-by-size": silver_members_by_size,
        "bronze-members-by-size": bronze_members_by_size,
        "unclassified-members-by-size": unclassified_members_by_size,
        "total-items-preserved": total_items_preserved,
        "total-items-unpreserved": total_items_unpreserved,
        "total-preservation-instances": total_preservation_instances,
        "total-items-sampled": total_items_sampled,
        "gold-member-ids": gold_member_ids,
        "silver-member-ids": silver_member_ids,
        "bronze-member-ids": bronze_member_ids,
        "unclassified-member-ids": unclassified_member_ids,
    }

    key = f"{REPORT_PATH}/{REPORT_FILENAME}"

    push_json_to_s3(
        s3=s3,
        json_obj=overall_result,
        bucket=REPORT_BUCKET,
        path=key,
    )

    return overall_result


def member_position(members=None, limit=0, member_id=None):
    import boto3
    import logging

    s3client = boto3.client("s3")
    session = boto3.Session()
    s3 = session.resource("s3")

    if not members:
        # get member list from S3
        members = get_members(
            s3client=s3client,
            samples_bucket=SAMPLES_BUCKET,
            samples_path=SAMPLES_PATH,
        )

    logging.info(
        f"Member {member_id} is at position {members.index(member_id)} of {len(members)} ({members.index(member_id) / len(members) * 100}%)"
    )


def rename_keys(members=None, limit=0, member_id=None):
    import boto3
    import logging

    s3client = boto3.client("s3")
    session = boto3.Session()
    s3 = session.resource("s3")

    if not members:
        # get member list from S3
        members = get_members(
            s3client=s3client,
            samples_bucket=SAMPLES_BUCKET,
            samples_path=SAMPLES_PATH,
        )

    for member in members:
        member_annotation = get_annotation(
            s3client,
            member,
            ANNOTATION_BUCKET,
            ANNOTATION_PATH,
            "preservation.json",
        )

        new_annotation = {}

        for key, value in member_annotation.items():
            new_key = normalize_json_key(key)
            new_annotation[new_key] = value

        key = f"{ANNOTATION_PATH}/members/{member}/preservation.json"

        push_json_to_s3(
            s3=s3,
            json_obj=new_annotation,
            bucket=ANNOTATION_BUCKET,
            path=key,
        )

        logging.info(f"Renamed keys for {member} in {key}")


def rescore_all(members=None, limit=0):
    import boto3
    from rich.progress import track

    s3client = boto3.client("s3")
    session = boto3.Session()
    s3 = session.resource("s3")

    if not members:
        # get member list from S3
        members = get_members(
            s3client=s3client,
            samples_bucket=SAMPLES_BUCKET,
            samples_path=SAMPLES_PATH,
        )

    if limit > 0:
        members = members[:limit]

    from joblib import Parallel, delayed

    results = Parallel(n_jobs=PARALLEL_JOBS)(
        delayed(rescore_member)(
            ANNOTATION_BUCKET,
            ANNOTATION_FILENAME,
            ANNOTATION_PATH,
            SAMPLES_BUCKET,
            SAMPLES_PATH,
            member_id,
            CODE_BUCKET,
        )
        for member_id in track(members)
    )


def get_annotation(
    s3client,
    member_id,
    annotations_bucket,
    annotations_path,
    annotation_name,
    path="members",
    use_member_id=True,
):
    if s3client is None:
        s3client = boto3.client("s3")

    if use_member_id:
        key = f"{annotations_path}/{path}/{member_id}/{annotation_name}"
    else:
        key = f"{annotations_path}/{path}/{annotation_name}"

    data = (
        s3client.get_object(Bucket=annotations_bucket, Key=key)["Body"]
        .read()
        .decode("utf-8")
    )

    from io import StringIO
    import json

    with StringIO(data) as json_file:
        output = json.loads(json_file.read())

        return output



import csv
import gc
import logging
import re
import sys
import tracemalloc
from csv import DictReader
from io import StringIO
from random import randint

import django
import requests
from crossref.restful import Etiquette, Journals, Works
from django.db import transaction
from django.db.models import QuerySet, Count

from lxml import etree as ET


def expand_issns(issn) -> set:
    """
    Expands a list of ISSNs using the ISSNL lookup database.
    :param issn: the list of ISSNs to expand
    :return: a set of expanded ISSNs
    """
    from preservationdatabase.models import ISSN, ISSNL

    final_issn_list = []

    for issn_item in issn:
        final_issn_list.append(issn_item)
        issn_objs = ISSN.objects.filter(identifier=issn_item)
        issnl_objs = ISSNL.objects.filter(identifier=issn_item)

        for issn_obj in issn_objs:
            issnl_objs = ISSNL.objects.filter(
                identifier=issn_obj.issnl.identifier
            )

            for issnl_obj in issnl_objs:
                issn_reversed = ISSN.objects.filter(issnl=issnl_obj)

                for issn_reversed_obj in issn_reversed:
                    final_issn_list.append(issn_reversed_obj.identifier)

        for issn_reversed in issnl_objs:
            final_issn_list.append(issn_reversed.identifier)

            issn_objs = ISSN.objects.filter(issnl=issn_reversed)

            for issn_obj in issn_objs:
                final_issn_list.append(issn_obj.identifier)

    final_issn_list = set(final_issn_list)

    return final_issn_list


def show_preservation(
    container_title: str,
    issn: set,
    volume: str,
    no: str | None,
    doi: str,
    archive: str = None,
    year: str | None = None,
    no_issn: bool = False,
    verbose: bool = False,
) -> (dict | None, str):
    """
    Determine whether an item is preserved
    :param container_title: the journal/container name
    :param issn: the ISSN
    :param volume: the volume
    :param no: the number
    :param doi: the DOI
    :param no_issn: if true, will not look up ISSNs for expansion
    :param year: the year
    :param verbose: if true, will print status updates
    :param archive: the archive to query (or None for all archives)
    :return: a dictionary of preservations and a doi
    """
    from preservationdatabase import constants

    # extend ISSNs
    issn = [] if not issn else issn
    issn = set(issn)

    # extend the ISSN using the ISSNL lookup table
    if no_issn:
        final_issn_list = issn
    else:
        final_issn_list = expand_issns(issn)

    if issn != final_issn_list and verbose:
        logging.info(
            f"Expanded ISSN list for title from {issn} to {final_issn_list}"
        )

    if archive is None:
        preservation_systems = [*constants.archives.values()]
    else:
        preservation_systems = [constants.archives[archive]]

    preservations = {}

    for system in preservation_systems:
        preserved, done = system.preservation(
            container_title, final_issn_list, volume, no=no, year=year, doi=doi
        )
        preservations[system.name()] = preserved, done

    return preservations, doi


def unpack_range(s: str) -> list:
    """Converts a range of numbers to a full set"""
    r = []
    for i in s.split(","):
        if "-" not in i and " to " not in i:
            r.append(int(i))
        elif " to " in i:
            l, h = map(int, i.split(" to "))
            r += range(l, h + 1)
        else:
            l, h = map(int, i.split("-"))
            r += range(l, h + 1)
    return r


def show_preservation_for_doi(
    doi_input: str, archive: str = None, no_issn: bool = False
) -> (dict | None, str):
    """
    Determine whether a DOI is preserved with resolution via the REST API
    :param doi_input: the DOI to look up
    :param no_issn: if true, will not look up ISSNs for expansion
    :param archive: the archive to query (or None for all archives)
    :return:
    """

    my_etiquette = Etiquette(
        "Preservation Status", "0.01", "https://eve.gd", "meve@crossref.org"
    )

    works = Works(etiquette=my_etiquette)
    doi = works.doi(doi_input)

    if not doi:
        logging.info(f"Unable to resolve DOI: {doi_input}")
        return {}, doi_input

    container_title = (
        doi["container-title"] if "container-title" in doi else None
    )
    issns = set(doi["ISSN"]) if "ISSN" in doi else None

    volume = doi["volume"] if "volume" in doi else None
    no = doi["issue"] if "issue" in doi else None

    year = None
    # "published": {"date-parts": [[2005, 12, 30]]}}}

    if "published" in doi:
        year = doi["published"]["date-parts"][0][0]

    logging.info(
        f'Checking {doi["DOI"]}: {container_title} '
        f"({issns}) v{volume} n{no} ({year})"
    )

    preserves, doi_echo = show_preservation(
        container_title,
        issns,
        volume,
        no,
        doi,
        archive,
        year=year,
        no_issn=no_issn,
    )

    # the "archive" field is apparently returned if deposited

    archives = set(doi["archive"]) if "archive" in doi else set()

    if "archive" in doi:
        for archive in archives:
            if archive in preserves and preserves[archive][0]:
                logging.info(
                    f"{doi_input} correctly asserts preservation in {archive}."
                )
            else:
                logging.info(
                    f"{doi_input} incorrectly asserts preservation in {archive}."
                )

    else:
        logging.info(f"{doi_input} asserts no preservation information.")

    return preserves, doi_input


def normalize_doi(doi: str) -> str:
    """
    Normalize a DOI
    :param doi: the DOI to normalize
    :return: a DOI without the prefix
    """

    # extract the DOI from the input
    # note that this is not as rigorous as it could be, but writing a single
    # expression that captures everything is hard.
    # See: https://www.crossref.org/blog/dois-and-matching-regular-expressions/
    pattern = r"(10.\d{4,9}/[-._;()/:A-Z0-9]+)"

    result = re.search(pattern, doi, re.IGNORECASE)

    return result.group(0) if result else None


def generic_lockss_import(
    url: str, model, skip_first_line: bool = False, local: bool = False
) -> None:
    """
    The generic import function for LOCKSS-like models
    :param url: the URL to download
    :param model: the model class to use
    :param local: whether to use a local file
    :param skip_first_line: whether to skip the first line of the file
    :return: None
    """
    from preservationdatabase.models import Publisher

    # get CSV data
    csv_file = download_remote(local, model, url)

    # clear out
    clear_out(model)

    with StringIO(csv_file) as input_file:
        # skip the top line in the CSV which is something like
        # #Keepers CLOCKSS serials 2022-12-19
        if skip_first_line:
            next(input_file)

        csv_reader = DictReader(input_file)

        for row in csv_reader:
            # trim the publisher field if needed
            publisher_name = row["Publisher"][:254]

            publisher, created = Publisher.objects.get_or_create(
                name=publisher_name
            )

            if (
                model.name() == "CLOCKSS"
                or model.name() == "LOCKSS"
                or model.name() == "Cariniana"
            ):
                # create the item
                model.create_preservation(
                    issn=row["ISSN"],
                    eissn=row["eISSN"],
                    title=row["Title"],
                    preserved_volumes=row["Preserved Volumes"],
                    preserved_years=row["Preserved Years"],
                    in_progress_volumes=row["In Progress Volumes"],
                    in_progress_years=row["In Progress Years"],
                    publisher=publisher,
                    model=model,
                )
            elif model.name() == "PKP PLN":
                model.create_preservation(
                    issn=row["ISSN"],
                    title=row["Title"],
                    preserved_volumes=row["Vol"],
                    preserved_no=row["No"],
                    publisher=publisher,
                    model=model,
                )

            logging.info(f'Added {row["Title"]} to {model.name()} data')


def clear_out(model):
    if hasattr(model, "name"):
        logging.info(f"Clearing previous {model.name()} data")
    model.objects.all().delete()


def download_remote(
    local,
    model,
    url,
    bucket="",
    s3client=None,
    decode=True,
    file=False,
    filename="",
):
    if s3client:
        if not file:
            if decode:
                return (
                    s3client.get_object(Bucket=bucket, Key=url)["Body"]
                    .read()
                    .decode("utf-8")
                )
            else:
                return s3client.get_object(Bucket=bucket, Key=url)[
                    "Body"
                ].read()
        else:
            logging.info("Storing downloaded file as {}".format(filename))
            s3client.download_file(bucket, url, filename)
            return filename

    if not local:
        logging.info(f"Downloading: {model.name()} data")
        csv_file = requests.get(url).text
    else:
        logging.info(f"Using local file for {model.name()} data")

        if decode:
            with open(url, "r") as f:
                csv_file = f.read()
        else:
            with open(url, "rb") as f:
                csv_file = f.read()

    return csv_file


def preservation_status(
    model, container_title, issn, volume, no=None, year=None
) -> (dict | None, str):
    """
    Determine whether a DOI is preserved in model
    :param model: the model class to use
    :param container_title: the container title
    :param issn: the ISSN
    :param volume: the volume
    :param no: the issue number
    :return: A model item (or None) and a bool indicating whether the item is
    fully preserved
    """
    preserved_item = get_preserved_item_record(model, container_title, issn)

    if not preserved_item or len(preserved_item) == 0:
        return None, False

    for pi in preserved_item:
        vols = [x.strip() for x in pi.preserved_volumes.split(";")]
        vols_in_prog = [x.strip() for x in pi.in_progress_volumes.split(";")]

        volume = str(volume)

        if volume in vols:
            return preserved_item, True
        elif volume in vols_in_prog:
            return preserved_item, False

        if year:
            years = [x.strip() for x in pi.preserved_years.split(";")]
            years_in_prog = [x.strip() for x in pi.in_progress_years.split(";")]

            if str(year) in years:
                return preserved_item, True
            elif str(year) in years_in_prog:
                return preserved_item, False

    return None, False


def get_preserved_item_record(model, container_title, issn) -> QuerySet | None:
    """
    Retrieves preservation records from the model
    :param model: the preservation model to use
    :param container_title: the name of the container
    :param issn: a list of ISSNs
    :return: a queryset of preservation model records or None
    """
    fields = [f.name for f in model._meta.get_fields()]

    # test ISSN
    try:
        if issn and "issn" in fields:
            preserved_item = None

            for sub_issn in issn:
                preserved_item = model.objects.filter(issn=sub_issn)
                if len(preserved_item) > 0:
                    break

            if not preserved_item or len(preserved_item) == 0:
                raise model.DoesNotExist
        else:
            raise model.DoesNotExist
    except model.DoesNotExist:
        # test EISSN
        try:
            if issn and "eissn" in fields:
                preserved_item = None

                for sub_issn in issn:
                    preserved_item = model.objects.filter(eissn=sub_issn)
                    if len(preserved_item) > 0:
                        break

                if not preserved_item or len(preserved_item) == 0:
                    raise model.DoesNotExist
            else:
                raise model.DoesNotExist
        except model.DoesNotExist:
            # test container title
            try:
                if container_title and "container_title" in fields:
                    preserved_item = model.objects.filter(title=container_title)
                    if not preserved_item or len(preserved_item) == 0:
                        raise model.DoesNotExist
                else:
                    raise model.DoesNotExist
            except model.DoesNotExist:
                return None

    return preserved_item


def process_onix(xml_file, elements, callback) -> None:
    """
    A faster method for processing ONIX XML than using DOM methods
    """
    current_object = {"volumes": [], "issues": []}

    collect = False

    count = 0

    for event, elem in ET.iterparse(xml_file, events=("start", "end")):
        if (
            event == "end"
            and elem.tag == "{http://www.editeur.org/onix/"
            "serials/SOH}HoldingsRecord"
        ):
            callback(current_object)

            gc.collect()

            current_object = {"volumes": [], "issues": []}

            count = count + 1
            stats = []

            filters = [tracemalloc.Filter(inclusive=True, filename_pattern="*")]

            if count % 100 == 0:
                django.db.reset_queries()

            # free memory (urgh)
            # a note: without this, the elementree will simply grow in memory
            # and previously iterated elements will not be garbage collected
            elem.clear()

        # detect if it's an ISSN
        if (
            event == "start"
            and elem.tag == "{http://www.editeur.org/onix/"
            "serials/SOH}ResourceVersionIDType"
            and elem.text == "07"
        ):
            collect = True
        elif (
            event == "start"
            and elem.tag == "{http://www.editeur.org/onix/"
            "serials/SOH}ResourceVersionIDType"
            and elem.text != "07"
        ):
            collect = False

        if (
            event == "start"
            and elem.tag == "{http://www.editeur.org/onix/"
            "serials/SOH}IDValue"
            and collect
        ):
            current_object["issn"] = normalize_issn(elem.text)

        if (
            event == "start"
            and elem.tag == "{http://www.editeur.org/onix/"
            "serials/SOH}TitleText"
            and collect
        ):
            current_object["title"] = elem.text

        if (
            event == "start"
            and elem.tag == "{http://www.editeur.org/onix/"
            "serials/SOH}PublisherName"
            and collect
        ):
            current_object["publisher"] = elem.text

        if (
            event == "start"
            and elem.tag == "{http://www.editeur.org/onix/"
            "serials/SOH}Coverage"
            and collect
        ):
            coverage = elem

            volume = ""
            issue = ""

            level_one = coverage.find(
                ".//{http://www.editeur.org/onix/serials/SOH}Level1"
            )
            level_two = coverage.find(
                ".//{http://www.editeur.org/onix/serials/SOH}Level2"
            )

            levels = [level_one, level_two]
            unit = None

            for level in levels:
                if level:
                    unit = level.find(
                        ".//{http://www.editeur.org/onix/serials/SOH}Unit"
                    )

                    # no idea why, but this HAS to be an explicit comparison
                    # with None rather than a boolean comparison
                    unit = unit.text if unit is not None else None

                    number = level.find(
                        ".//{http://www.editeur.org/onix/serials/SOH}Number"
                    )

                    number = number.text if number is not None else None

                    if unit == "Volume":
                        volume = number
                    elif unit == "Issue":
                        issue = number
                    else:
                        print("Unrecognised unit: {}".format(unit))

            if unit:
                current_object["volumes"].append(volume)
                current_object["issues"].append(issue)

        if (
            event == "start"
            and elem.tag == "{http://www.editeur.org/onix/"
            "serials/SOH}PreservationStatusCode"
            and collect
        ):
            current_object["status"] = elem.text

        if (
            event == "start"
            and elem.tag == "{http://www.editeur.org/onix/"
            "serials/SOH}VerificationStatus"
            and collect
        ):
            current_object["verification"] = elem.text

    return


def normalize_issn(issn) -> str:
    """
    Normalizes an ISSN
    :param issn: the ISSN
    :return: the normalized ISSN
    """
    issn = "" if issn is None else issn.strip()

    return f"{issn[0:4]}-{issn[4:8]}" if issn and "-" not in issn else issn


@transaction.atomic
def random_db_entries(model, count, only_crossref=False) -> list:
    """
    Returns random entries from a database model
    :param model: the model to use
    :param count: the number of random entries to return
    :param only_crossref: whether to only return crossref entries
    :return: a list of random entries
    """

    # NB we do it like this because queryset.order_by('?') can be incredibly
    # slow
    output = []

    total = model.objects.all().count()

    my_etiquette = Etiquette(
        "Preservation Status", "0.01", "https://eve.gd", "meve@crossref.org"
    )

    jrnls = Journals(etiquette=my_etiquette)

    while len(output) < count - 1:
        random_index = randint(0, total - 1)

        entry = model.objects.all()[random_index]

        if only_crossref:
            append = False

            new_issns = expand_issns([entry.issn])

            for issn in new_issns:
                jrnl = jrnls.journal(issn=issn)

                if jrnl:
                    append = True
                    break

            if hasattr(entry, "eissn"):
                new_eissns = expand_issns([entry.eissn])

                for issn in new_eissns:
                    jrnl = jrnls.journal(issn=issn)

                    if jrnl:
                        append = True
                        break

            if append:
                output.append(entry)
                print(
                    "Found crossref entry ({}): {}".format(len(output), entry)
                )
        else:
            output.append(entry)

    return output


def in_archive(item, archive) -> bool:
    """
    Determines whether a container is in an archive
    :param item: the container to check
    :param archive: the archive to check
    :return: True if both items are in the same archive, False otherwise
    """

    fields = item._meta.fields

    if "issn" in fields:
        issns = archive.objects.filter(issn=item["issn"])

        if len(issns) > 0:
            return True

    if "eissn" in fields:
        try:
            issns = archive.objects.filter(eissn=item["eissn"])

            if len(issns) > 0:
                return True
        except:
            issns = archive.objects.filter(issn=item["eissn"])

            if len(issns) > 0:
                return True

    if "title" in fields:
        titles = archive.objects.filter(title=item["title"])

        if len(titles) > 0:
            return True

    return False


def import_issnl(issnl_file, issnl_model, issn_model):
    from rich.progress import track

    csv.field_size_limit(sys.maxsize)

    clear_out(issn_model)
    clear_out(issnl_model)

    logging.info("Erased existing ISSN-ISSNL mappings.")

    with open(issnl_file, "r") as input_file:
        csv_reader = csv.DictReader(input_file, delimiter="\t")

        for row in track(csv_reader):
            if row["ISSN"] != row["ISSN-L"]:
                issnl, created = issnl_model.objects.get_or_create(
                    identifier=row["ISSN-L"]
                )

                issn, created = issn_model.objects.get_or_create(
                    identifier=row["ISSN"], issnl=issnl
                )

                logging.info(
                    f"Created ISSN-ISSNL mapping: {issn.identifier} {issnl.identifier}"
                )



from preservationdatabase.models import (
    CarinianaPreservation,
    ClockssPreservation,
    HathiPreservation,
    InternetArchivePreservation,
    InternetArchiveItem,
    LockssPreservation,
    OculScholarsPortalPreservation,
    PKPPreservation,
    PorticoPreservation,
)

archives = {
    "cariniana": CarinianaPreservation,
    "clockss": ClockssPreservation,
    "hathitrust": HathiPreservation,
    "internet_archive": InternetArchivePreservation,
    "internet_archive_scholar": InternetArchiveItem,
    "lockss": LockssPreservation,
    "pkp": PKPPreservation,
    "portico": PorticoPreservation,
    "ocul": OculScholarsPortalPreservation,
}



import json
import pathlib
import unittest
import logging
from preservationdatabase import constants, utils

logger = logging.getLogger(__name__)
logging.disable(logging.NOTSET)
logger.setLevel(logging.DEBUG)


class ArchiveTestCase(unittest.TestCase):
    """
    Tests for archival presences and absences using known good data
    """
    def test(self):
        test_file = \
            pathlib.Path(
                pathlib.Path(__file__).resolve().parent.parent / 'test_data'
                / 'tests.jsonl')
        logging.debug(f'Loading tests from {test_file}')

        with open(test_file, 'r') as json_file:
            json_list = list(json_file)

        for json_str in json_list:
            result = json.loads(json_str)

            # resolve archive object
            archive = result['archive']
            archive_object = constants.archives[archive]
            doi = result['doi']
            should_be_present = result['present']

            preservation_status, doi = \
                utils.show_preservation_for_doi(doi, archive)

            if should_be_present:
                print(f'Testing if {doi["DOI"]} is in {archive}')
                logging.debug(f'Testing if {doi["DOI"]} is in {archive}')

                self.assertTrue(archive_object.name() in preservation_status,
                                msg=f'{doi["DOI"]} is not in {archive}')

                if archive_object.name() in preservation_status:
                    self.assertFalse(
                        preservation_status[
                            archive_object.name()] == (None, False),
                        msg=f'{doi["DOI"]} is not in {archive}')
            else:
                print(f'Testing if {doi["DOI"]} is not in {archive}')
                logging.debug(f'Testing if {doi["DOI"]} is not in {archive}')

                self.assertTrue(
                    archive_object.name() not in preservation_status or
                    preservation_status[archive_object.name()] == (None, False))


if __name__ == '__main__':
    unittest.main()



import logging


def setup_environment(
    code_bucket, download_settings=True, verbose=False
) -> None:
    """
    Initialize the Django ORM environment.
    :param code_bucket: the S3 bucket where the code and settings are stored.
    :param download_settings: whether to download the settings file from S3.
    :param verbose: whether to log to the console.
    :return: None
    """
    import os
    import sys
    import inspect
    from pathlib import Path

    from rich.console import Console
    from rich.logging import RichHandler

    logging.basicConfig(
        level=logging.INFO,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(console=Console(stderr=True))],
    )
    log = logging.getLogger("preservation-database")
    log.setLevel(logging.INFO)

    logging.getLogger("boto3").setLevel(logging.WARNING)
    logging.getLogger("botocore").setLevel(logging.WARNING)

    current_directory = os.path.dirname(
        os.path.abspath(inspect.getfile(inspect.currentframe()))
    )

    if download_settings:
        settings_file = Path(current_directory) / "downloaded_settings.py"

        if verbose:
            logging.info(
                "Settings file downloading to: " "{}".format(settings_file)
            )

        import boto3

        s3client = boto3.client("s3")
        s3client.download_file(code_bucket, "settings.py", str(settings_file))

    if code_bucket:
        sys.path.append(current_directory)
        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "downloaded_settings")

    import django

    django.setup()
